import path from 'path'
import yargs from 'yargs-parser'
import fs from 'fs-extra'
import type { ParamsResult, RunnerConfig } from './types'

import prompt from './prompt'
export const DEFAULT_ACTION = '_default'

const resolvePositionals = async (templates: string, args: string[]) => {
  /*
  we want a to create flexible resolution and allow both:

  1. [gen] [action] [NAME]
  2. [gen] [NAME] which, if NAME=action, has a conflict, so goes to the generator with
      an empty name, otherwise -- goes to the new '_default' generator, with a name of 'NAME'

  E.g.
  for a template layout of:
  init/
    repo/
    ..

  init repo (repo exists, so goes to the repo gen, this is not a default named 'repo'!)
  init repo MyName
  init MyName (default, name=MyName), default because 'repo' does not exist
  init (default, name=[empty]), default always!
  */
  let [generator, action, name] = args
  if (
    generator &&
    action &&
    (await fs.exists(path.join(templates, generator, action)))
  ) {
    return [generator, action, name]
  }

  if (
    generator &&
    (await fs.exists(path.join(templates, generator, DEFAULT_ACTION)))
  ) {
    action = DEFAULT_ACTION
    ;[generator, name] = args
  }
  return [generator, action, name]
}

const params = async (
  { templates, createPrompter }: RunnerConfig,
  externalArgv: string[],
): Promise<ParamsResult> => {
  const argv = yargs(externalArgv)

  // XXX: inline
  // tricky. need to also do listing in help, and others based on inline and not fs.
  // need to map out every aspect that ALSO needs inline operations.
  // need to decide if this is inline or fs based templates.
  // if inline, grab from json data, run prompts, and return.
  /*
  1. OK: split render to read files and render files with a contract of a file between.
  this is harmless, and also opens up for "just the engine".
  2. OK: split prompting and fusion with args to clean prompt(questions, args)
  so that it can be preloaded from disk or inmem
  3. split prompt to 'run hooks from hook file' and 'run prompt with data' which is given or
  not given from hooks file
  4. execute-inline: engine is responsible to take argv, turn it into gen tuple,
  decides if exists or not, then runs params (prompting) and then render.
      1. it has runner config, so first, check in there instead of FS
      2. if inline, run prompts from inline and not disk
        prompts is the most tricky part to split
        inline-prompt.ts and hooks-prompt.ts ?
      3. run render from inline, and not disk
      4. fix 'help' to show from inline as well as disk

  */

  const [generator, action, name] = await resolvePositionals(templates, argv._)

  if (!generator || !action) {
    return { generator, action, templates }
  }
  const [mainAction, subaction] = action.split(':')

  /*
  this is how to do inline generators via .hygen.js
  RunnerConfig
   "generators": {
      "models":{ <------ mainAction
        "_default":{ <------ subaction
          prompts: [
            ..
          ],
          templates: [
            {
              file: ".."
              attributes: {}
              body: {}
            }
          ]
        }
      }
    }
  */
  const actionfolder = path.join(templates, generator, mainAction)

  const { _, ...cleanArgv } = argv
  const promptArgs = await prompt(createPrompter, actionfolder, {
    // NOTE we might also want the rest of the generator/action/etc. params here
    // but theres no usecase yet
    ...(name ? { name } : {}),
    ...cleanArgv,
  })

  const args = Object.assign(
    {
      templates,
      actionfolder,
      generator,
      action,
      subaction,
    },
    { _ }, // include positionals as special arg for templates to consume
    cleanArgv,
    name && { name },
    promptArgs,
  )

  return args
}

export default params
